/*
**  DrawIt converter for 3Com PalmPilots
**  main.cpp
**
**  Copyright (C) 2001  William McFarland
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
** 
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
** 
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
**  I basically "borrowed" the sketelon of this program from txt2pdbdoc
*/

/* standard */
#include "drawit2ppm.h"
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <iostream>
#include <assert.h>

using namespace std;

/*****************************************************************************
 *
 * SYNOPSIS
 *
 *
 * DESCRIPTION
 *
 *  Parse the command line, initialize, call other functions ... the
 *  usual things that are done in main().
 *
 * PARAMETERS
 *
 *  argc    The number of arguments.
 *
 *  argv    A vector of the arguments; argv[argc] is null.  Aside from
 *      the options below, the arguments are the names of the files.
 *
 * SEE ALSO
 *
 *  Brian W. Kernighan, Dennis M. Ritchie.  "The C Programming Language,
 *  2nd ed."  Addison-Wesley, Reading, MA.  pp. 114-118.
 *
 *****************************************************************************/

int main( int argc, char *argv[] )
{
    extern char*    optarg;
    extern int      optind, opterr;

    bool            get_num = false;
    int             img_num = 0;
    
    bool            use_basename = false;
    char            basename [BUFF_SIZE];
    
    char const      opts[] = "b:n:vV"; /* command line options */
    int             opt;            /* option being processed */

    // Set the default basename
    memset(basename, 0, BUFF_SIZE);
    sprintf(basename, "drawit_image");

    // Set the name of the program
    me = strrchr( argv[0], '/' );       /* determine base name... */
    me = me ? me + 1 : argv[0];         /* ...of executable */

    select_swap();

    /********** Process command-line options *****************************/

    opterr = 1;
    while ( (opt = getopt( argc, argv, opts )) != EOF )
        {   
        switch ( opt ) 
            {
            case 'n':
                get_num = true;
                img_num = strtol (optarg, (char **)NULL, 10);
                break;

            case 'b':
                use_basename = true;
                if (strlen(optarg) < BUFF_SIZE)
                {
                    memset(basename, 0, BUFF_SIZE);
                    sprintf(basename, "%s", optarg);
                }
                else
                {
                    cout << "Basename too long, name limited to " << BUFF_SIZE << " characters\n";
                    usage();
                }
                break;

            case 'v':
                verbose = true;
                break;
                
            case 'V':
                printf( "%s %s\n", me, VERSION );
                exit( 0 );

            case '?': /* Bad option. */
                usage();
            }
        }

    argc -= optind, argv += optind;

    if (argc != 1)
    {
        usage();
    }
    else if ( get_num )
    {
        extract(argv[0], img_num, basename);
    }
    else
    {
        extract_all(argv[0], basename);
    }

    exit( 0 );
}

/*****************************************************************************
 *
 * SYNOPSIS
 *
 * DESCRIPTION
 *
 *  Determine the "endianness" of the machine by copying byte strings
 *  into a union and seeing how they come out when looked at as integers;
 *  set pointers-to-function appropriately.
 *
 *****************************************************************************/
void select_swap()
{
    union { char c[2];  Word n; }  w;
    union { char c[4]; DWord n; } dw;

    strncpy(  w.c, "\1\2",     2 );
    strncpy( dw.c, "\1\2\3\4", 4 );

    swap_Word  =     w.n == 0x0102 ? swap_Word_12 :
             w.n == 0x0201 ? swap_Word_21 : 0;

    swap_DWord =    dw.n == 0x01020304 ? swap_DWord_1234 :
            dw.n == 0x04030201 ? swap_DWord_4321 : 0;

    if ( swap_Word && swap_DWord )
        return;

    fprintf( stderr, "%s: couldn't select byte-swapping algorithm\n", me );
    exit( 2 );
}  

void usage(void)
{
    fprintf( stderr, "\
usage: %s [-n #] [-V] [-b name] DrawItDB.pdb\n\n\
 options:\n\
 --------\n\
  -n: number of image to extract (1 to total images)\n\
  -b: Use this as the basename for the images
  -v: verbose mode
  -?: prints the usage message
  -V: print version informaiton\n", me, me );
    exit( 1 );
}

Word swap_Word_12( Word r ) {
    return r;  
}

Word swap_Word_21( Word r ) {
    return (r >> 8) | (r << 8);
}

DWord swap_DWord_1234( DWord r ) {
    return r;
}

DWord swap_DWord_4321( DWord r ) {
    return  ( (r >> 24) & 0x00FF ) |   (r << 24) |
        ( (r >>  8) & 0xFF00 ) | ( (r <<  8) & 0xFF0000 );
}

FILE* open_src_file( char const *file_name ) {
    FILE *f = fopen( file_name, "rb" );
    if ( f ) return f;
    fprintf( stderr, "%s: can not open %s for input\n", me, file_name );
    exit( 2 );
}

FILE* open_dest_file( char const *file_name )
{
    FILE *f = fopen( file_name, "wb" );
    if ( f )
    {
        return f;
    }
    fprintf( stderr, "%s: can not open %s for output\n", me, file_name );
    exit( 2 );
}

void extract_all (char const *infile, char const * basename)
{
    di_image * img = NULL;
    
    int num_records = load_data(infile, img);

    assert (img != NULL);
    
    // Write out the images in ppm format
    for (int cnt=0; cnt<num_records; cnt++)
    {
        // Write the ppm file
        img[cnt].write_image(basename);
        if (verbose)
        {
            cout << "Wrote image # " << cnt+1 << " using a basename of " << basename << eol;
        }
        
    }
    // Cleanup the allocated memory
    delete [] img;
    img = NULL;
    return;
}

void extract (char const *infile, int img_num, char const * basename)
{
    di_image * img = NULL;
    
    int num_records = load_data(infile, img);

    assert (img != NULL);
    
    // Write the ppm file
    int inum = (img_num - 1);
    if (inum < 0)
    {
        // Most likely someone counting from zero
        img[img_num].write_image("drawit_image");
        cout << "Incorrect usage, image numbers start at 1, I extracted the first image\n";
        usage ();
    }
    else if(inum >= num_records)
    {
        // Whoa! I guess they forgot how many images there are in the database,
        // give them the last one
        img[num_records-1].write_image(basename);
        cout << "Image number too large.  I extracted the last image.\n";
        usage ();
    }
    else
    {
        // Correct range.
        img[inum].write_image(basename);
    }
    if (verbose)
    {
        cout << "Wrote image # " << img_num << " using a basename of " << basename << eol;
    }
    
    // Cleanup the allocated memory
    delete [] img;
    img = NULL;

    return;
}

//
// We will load the data from the database.
// This function returns the number of records and also has the pointer
// point to the image array.
//
// Precondition: img = NULL;
//
// WARNING: the calling function must free the memory
//
int load_data (char const *infile, di_image * &img)
{
    assert(NULL == img);

    FILE               *fin;
    pdb_header         header;
    DWord              offset;
    
    // Open the input file
    fin = open_src_file (infile);
    
    // Read the header
    fread (&header, PDB_HEADER_SIZE, 1, fin);
    
    // Verify we have the correct file
    if (strncasecmp (header.type, TYPE, sizeof(header.type)) ||
        (strncmp (header.creator, CREATOR, sizeof(header.creator))))
    {
        cerr << me << ": " << infile << " is not a DrawIt archive" << eol;
        exit (3);
    }
    
    int num_records = swap_Word (header.numRecords);
    if (verbose)
    {
        cout << me << ": Number of records: " << num_records << eol;
    }

    // How big is the file
    fseek( fin, 0, SEEK_END );
    long file_size = ftell( fin );
    
    // Populate the record vector
    fseek(fin, PDB_HEADER_SIZE, SEEK_SET);
    drawit_record * rec = new drawit_record[num_records];
    img = new di_image[num_records];
    for (int cnt=0; cnt<num_records; cnt++)
    {
        fread(&rec[cnt], PDB_RECORD_HEADER_SIZE, 1, fin);
        int  offset = (long)swap_DWord (rec[cnt].offset);
        img[cnt].set_image_number(cnt);
        img[cnt].set_offset(offset);
        if (verbose)
        {
            cout << "Offset = " << offset << eol;
        }
        if (cnt != 0)
        {
            long prev_offset = img[cnt-1].get_offset();
            img[cnt-1].set_size(offset - prev_offset);
        }
    }
    // Calculate the last record size
    img[num_records-1].set_size(file_size - img[num_records-1].get_offset());
    
    // Read each image and output it in ppm format
    for (int cnt=0; cnt<num_records; cnt++)
    {
        // Read in the image from the database file
        img[cnt].read_buffer(fin);
    }

    // Clean up what memory we can, the calling function must cleanup
    // the image array.
    delete [] rec;
    rec = NULL;

    assert (img != NULL);
    return num_records;
}
